import logging

from typing import cast

import numpy as np
import pytensor
import pytensor.tensor as pt

from pytensor.compile import Supervisor, mode
from pytensor.compile.builders import OpFromGraph
from pytensor.graph import FunctionGraph
from sympytensor import as_tensor

from cge_modeling.tools.parsing import normalize_eq
from cge_modeling.tools.pytensor_tools import (
    at_least_list,
    clone_and_rename,
    flatten_equations,
    get_required_inputs,
    make_jacobian,
    make_jacobian_from_sympy,
    make_printer_cache,
    object_to_pytensor,
    rewrite_pregrad,
    unpacked_graph_to_packed_graph,
)

_log = logging.getLogger(__name__)


def pytensor_objects_from_CGEModel(cge_model):
    n_eq = len(cge_model.unpacked_variable_names)
    variables = [object_to_pytensor(var, cge_model.coords) for var in cge_model.variables]
    parameters = [object_to_pytensor(param, cge_model.coords) for param in cge_model.parameters]

    cache = make_printer_cache(variables, parameters)
    unpacked_cache = {}
    _log.info("Converting equations to pytensor graphs")

    if cge_model.parse_equations_to_sympy:
        # This will be a flat list of equations with a unique variable for each unpacked object
        # We want to rewrite this graph so that it will accept the packed array inputs (held in the cache dictionary)
        pytensor_equations = [
            as_tensor(eq, cache=unpacked_cache) for eq in cge_model.unpacked_equation_symbols
        ]

        # We need to replace the indexed variables with the unpacked variables
        pytensor_equations = unpacked_graph_to_packed_graph(
            pytensor_equations, cge_model, cache, unpacked_cache
        )
        flat_equations = pt.stack(pytensor_equations)

    else:
        cache = {k[0]: v for k, v in cache.items()}
        cache.update({"pt": pt})

        # Copy the cache to call with eval, otherwise python will add a bunch of global environment variables to it
        eval_cache = cache.copy()
        pytensor_equations = [
            eval(normalize_eq(eq.equation), eval_cache) for eq in cge_model.equations
        ]
        flat_equations = flatten_equations(pytensor_equations)

    flat_equations = pt.specify_shape(flat_equations, (n_eq,))
    flat_equations.name = "equations"

    _log.info("Applying pre-gradient rewrites")
    rewrite_pregrad(flat_equations)

    return flat_equations, variables, parameters, (cache, unpacked_cache)


def compile_cge_model_to_pytensor(
    cge_model,
    sparse=False,
) -> tuple[tuple[list, list], tuple[pt.TensorLike, pt.TensorLike, pt.TensorLike]]:
    """
    Compile a CGE model to a PyTensor function.

    Parameters
    ----------
    cge_model: CGEModel
        The CGE model object to compile

    sparse: bool, optional
        Whether to use sparse matrices for the Jacobian and its inverse. Defaults to False.

    Returns
    -------
    inputs: tuple
        variables: list[pytensor.tensor.TensorVariable] of length cge_model.n_variables
            A list of pytensor variables representing the variables in the system of equations. The order of the
            variables is the same as the in which the user supplied them to the CGE model.

        parameters: list[pytensor.tensor.TensorVariable] of length cge_model.n_parameters
            A list of pytensor variables representing the parameters in the system of equations. The order of the
            parameters is the same as the in which the user supplied them to the CGE model.

    outputs: tuple
        flat_equations: pytensor.tensor.TensorVariable
            A single pytensor vector representing the equations in the system of equations. The length of the vector is
            equal to the number of *unrolled* equations in the system of equations.

        jac: pytensor.tensor.TensorVariable
            A pytensor matrix representing the Jacobian of the system of equations. The shape of the matrix is
            (n_eq, n_eq), where n_eq is the number of *unrolled* equations in the system of equations.

        B: pytensor.tensor.TensorVariable
            A pytensor matrix representing the Jacobian of the system of equations with respect to the parameters.
            The shape of the matrix is (n_eq, n_params), where n_eq is the number of *unrolled* equations in the system
            of equations and n_params is the number of *unrolled* parameters in the system of equations.
    """
    (
        flat_equations,
        variables,
        parameters,
        sympy_to_pytensor_caches,
    ) = pytensor_objects_from_CGEModel(cge_model)
    cache, unpacked_cache = sympy_to_pytensor_caches

    # n_eq = flat_equations.type.shape[0]
    inputs = (variables, parameters)

    if not all([x in cache.values() for x in inputs[0] + inputs[1]]):
        missing_inputs = set(inputs[0] + inputs[1]) - set(cache.values())
        raise ValueError(
            f"Cannot compile pytensor graph because the following inputs are missing: {missing_inputs}"
        )

    required_inputs = get_required_inputs(flat_equations)
    missing_inputs = set(required_inputs) - set(variables + parameters)
    extra_inputs = set(variables + parameters) - set(required_inputs)

    if missing_inputs:
        raise ValueError(
            f"Cannot compile pytensor graph because the following inputs are missing: {missing_inputs}"
        )
    if extra_inputs:
        raise ValueError(
            f"Cannot compile pytensor graph because the following inputs were provided but are not required: {extra_inputs}"
        )

    if cge_model.parse_equations_to_sympy:
        _log.info("Computing Jacobian")
        jac = make_jacobian_from_sympy(
            cge_model,
            wrt="variables",
            sparse=sparse,
            cache=cache,
            unpacked_cache=unpacked_cache,
        )

        jac_inputs = get_required_inputs(jac)
        assert all([x in cache.values() for x in jac_inputs])

        _log.info("Computing B matrix (derivatives w.r.t parameters)")
        B = make_jacobian_from_sympy(
            cge_model,
            wrt="parameters",
            sparse=sparse,
            cache=cache,
            unpacked_cache=unpacked_cache,
        )

    else:
        _log.info("Computing Jacobian")
        jac = make_jacobian(flat_equations, variables)

        _log.info("Computing B matrix (derivatives w.r.t parameters)")
        B = make_jacobian(flat_equations, parameters)

    jac.name = "jacobian"
    B.name = "B"

    # _log.info("Inverting jacobian")
    # if inverse_method == "pinv":
    #     jac_inv = pt.linalg.pinv(jac)
    # elif inverse_method == "solve":
    #     jac_inv = pt.linalg.solve(jac, pt.eye(jac.shape[0]), check_finite=False)
    # elif inverse_method == "svd":
    #     U, S, V = pt.linalg.svd(jac)
    #     S_inv = pt.where(pt.gt(S, 1e-8), 1 / S, 0)
    #     jac_inv = V @ pt.diag(S_inv) @ U.T
    # else:
    #     raise ValueError(
    #         f'Invalid inverse method {inverse_method}, expected one of "pinv", "solve", "svd"'
    #     )

    # jac_inv = pt.specify_shape(jac_inv, (n_eq, n_eq))
    # jac_inv.name = "inverse_jacobian"

    outputs = (flat_equations, jac, B)

    return inputs, outputs


def compile_cge_model_to_pytensor_Op(cge_model) -> tuple[pt.Op, pt.Op]:
    """
    Compile a CGE model to a PyTensor Ops.

    Parameters
    ----------
    cge_model: CGEModel
        The CGE model object to compile

    Returns
    -------
    f_model: pt.Op
        A PyTensor Op representing computation of the residuals of model equations given model variables and
        parameters as inputs

    f_jac: pt.Op
        A PyTensor Op representing computation of the Jacobian of model equations given model variables and
        parameters as inputs

    Notes
    -----
    In general, it shouldn't be necessary to use this function. Most downstream computation can directly use the graph
    generated by compile_cge_model_to_pytensor. In the case of optimization, however, the function and its Jacobian
    need to be "anonymous". This function exists to facilitate that use case.
    """

    (variables, parameters), outputs = compile_cge_model_to_pytensor(cge_model)

    flat_equations, jac, B = outputs
    inputs = list(variables) + list(parameters)

    f_model = OpFromGraph(inputs, outputs=[flat_equations], inline=True)
    f_jac = OpFromGraph(inputs, outputs=[jac], inline=True)

    return f_model, f_jac


def flat_tensor_to_ragged_list(tensor, shapes):
    out = []
    cursor = 0

    for shape in shapes:
        s = int(np.prod(shape))
        x = tensor[cursor : cursor + s].reshape(shape)
        out.append(x)
        cursor += s

    return out


def euler_approximation(
    A: pt.TensorVariable,
    B: pt.TensorVariable,
    variables: list[pt.Variable],
    parameters: list[pt.Variable],
    n_steps: int = 100,
):
    """
    Find the solution to a system of equations using the Euler approximation method.

    The Euler approximation method is a simple method for finding the solution to a system of equations. It is
    an extension of a first order Taylor approximation, improving on the linearized approximation by decomposing
    the move from the initial (known) solution, around which the system is linearized, to the final (unknown) point
    into a series of small steps. At each step, linear approximation is re-computed around the previous step. By this
    method, the solution to the system can be traced out to arbitrary precision by iteratively taking small steps.

    Parameters
    ----------
    A: pytensor.tensor.TensorVariable
        Inverse of the Jacobian of the system of equations with respect to the variables
    B: pytensor.tensor.TensorVariable
        Jacobian of the system of equations with respect to the parameters
    variables: list of pytensor.tensor.TensorVariable
        A list of pytensor variables representing the variables in the system of equations
    parameters: list of pytensor.tensor.TensorVariable
        A list of pytensor variables representing the parameters in the system of equations
    n_steps: int
        The number of steps to take in the Euler approximation

    Returns
    -------
    theta_final: pytensor.tensor.TensorVariable
        A symbolic tensor representing the full trajectory of parameter updates used to compute the solution to the
        system of equations at the final point, theta_final[-1].

    result: pytensor.tensor.TensorVariable
        The values of the variables at each step of the Euler approximation, corresponding to linear approximation
        around the point (x[i-1], theta_final[i-1]) at each step i.

    """
    x_list = cast(list, at_least_list(variables))
    theta_list = cast(list, at_least_list(parameters))
    theta_final = pt.concatenate(
        [pt.atleast_1d(clone_and_rename(x)).flatten() for x in theta_list], axis=-1
    )

    theta0 = flatten_equations(theta_list)

    dtheta = theta_final - theta0
    step_size = dtheta / n_steps

    Bv = B @ pt.atleast_1d(step_size)
    Bv.name = "Bv"

    step = -pt.linalg.solve(A, Bv, assume_a="gen", check_finite=False)

    f_step = OpFromGraph(x_list + theta_list + [step_size], [step], inline=True)

    x_args = len(x_list)
    x_shapes = [x.type.shape for x in x_list]
    theta_shapes = [theta.type.shape for theta in theta_list]

    def step_func(*args):
        x_prev = args[:x_args]
        theta_prev = args[x_args:-1]
        step_size = args[-1]

        step = f_step(*x_prev, *theta_prev, step_size)
        delta_x = flat_tensor_to_ragged_list(step, x_shapes)
        x_next = [x + dx for x, dx in zip(x_prev, delta_x)]

        delta_theta = flat_tensor_to_ragged_list(step_size, theta_shapes)
        theta_next = [theta + dtheta for theta, dtheta in zip(theta_prev, delta_theta)]

        assert len(theta_next) == len(theta_prev)
        return x_next + theta_next

    result, updates = pytensor.scan(
        step_func,
        outputs_info=x_list + theta_list,
        non_sequences=[step_size],
        n_steps=n_steps,
    )

    final_result = []
    for i, x in enumerate(x_list + theta_list):
        x_with_initial = pt.concatenate([pt.atleast_Nd(x, n=result[i].ndim), result[i]], axis=0)
        final_result.append(x_with_initial)

    return theta_final, final_result


def pytensor_euler_step(system, A, B, variables, parameters):
    x_list = at_least_list(variables)
    x_shapes = [x.type.shape for x in x_list]

    theta_list = at_least_list(parameters)

    n_steps = pt.iscalar("n_steps")
    theta_final = [x.clone(name=f"{x.name}_final") for x in theta_list]
    theta0 = [x.clone(name=f"{x.name}_initial") for x in theta_list]

    step_size = [(x - y) / n_steps for x, y in zip(theta_final, theta0)]

    try:
        Bv = pytensor.gradient.Rop(system, parameters, step_size)
    except NotImplementedError:
        _log.warning(
            "Tried to compute jvp of model with respect to parameters but failed; an Op in your model does not "
            "have an implemented Rop method. Falling back to direct computation via jacobian. This is slow and "
            "memory intensive for large models."
        )
        v = flatten_equations(step_size)
        Bv = B @ v

    step = -pt.linalg.solve(A, Bv)
    step.name = "euler_step"

    delta_x = flat_tensor_to_ragged_list(step, x_shapes)
    x_next = [x + dx for x, dx in zip(x_list, delta_x)]

    theta_next = [theta + dtheta for theta, dtheta in zip(theta_list, step_size)]

    inputs = x_list + theta_list + theta0 + theta_final + [n_steps]
    outputs = x_next + theta_next

    return inputs, outputs


def jax_loss_grad_hessp(system, variables, parameters):
    import jax
    import jax.numpy as jnp

    from pytensor.link.jax.dispatch import jax_funcify

    loss = (system**2).sum()
    fgraph = FunctionGraph(inputs=variables + parameters, outputs=[loss], clone=True)
    fgraph.attach_feature(
        Supervisor(
            input
            for input in fgraph.inputs
            if not (hasattr(fgraph, "destroyers") and fgraph.has_destroyers([input]))
        )
    )
    mode.JAX.optimizer.rewrite(fgraph)
    f_loss_jax = jax_funcify(fgraph)

    x_shapes = [x.type.shape for x in variables]
    theta_shapes = [x.type.shape for x in parameters]

    def f_loss_wrapped(x, theta):
        xs = flat_tensor_to_ragged_list(x, x_shapes)
        thetas = flat_tensor_to_ragged_list(theta, theta_shapes)

        return f_loss_jax(*xs, *thetas)[0]

    f_loss = jax.jit(f_loss_wrapped)

    grad = jax.grad(f_loss_wrapped, 0)

    def f_grad_jax(x, theta):
        return jnp.stack(grad(x, theta))

    f_grad = jax.jit(f_grad_jax)

    _f_hess_jax = jax.jacfwd(f_grad_jax, argnums=0)

    def f_hess_jax(x, theta):
        return jnp.stack(_f_hess_jax(x, theta))

    f_hess = jax.jit(f_hess_jax)

    def f_hessp_jax(x, p, theta):
        _, u = jax.jvp(lambda x: f_grad_jax(x, theta), (x,), (p,))
        return jnp.stack(u)

    f_hessp = jax.jit(f_hessp_jax)

    return f_loss, f_grad, f_hess, f_hessp


def jax_euler_step(system, variables, parameters):
    import jax
    import jax.numpy as jnp

    from pytensor.link.jax.dispatch import jax_funcify

    fgraph = FunctionGraph(inputs=variables + parameters, outputs=[system], clone=True)
    fgraph.attach_feature(
        Supervisor(
            input
            for input in fgraph.inputs
            if not (hasattr(fgraph, "destroyers") and fgraph.has_destroyers([input]))
        )
    )
    mode.JAX.optimizer.rewrite(fgraph)
    f_system_jax = jax_funcify(fgraph)

    x_shapes = [x.type.shape for x in variables]
    theta_shapes = [x.type.shape for x in parameters]

    variable_names = [x.name for x in variables]
    parameter_names = [x.name for x in parameters]

    def f_sys_wrapped(x, theta):
        xs = flat_tensor_to_ragged_list(x, x_shapes)
        thetas = flat_tensor_to_ragged_list(theta, theta_shapes)

        return f_system_jax(*xs, *thetas)[0]

    f_sys = jax.jit(f_sys_wrapped)

    def step(**kwargs):
        n_steps = kwargs.pop("n_steps")
        x_list = [kwargs.pop(x) for x in variable_names]

        theta_list = [kwargs.pop(x) for x in parameter_names]
        theta0 = [kwargs.pop(f"{x}_initial") for x in parameter_names]
        theta_final = [kwargs.pop(f"{x}_final") for x in parameter_names]

        x_vec = jnp.concatenate([jnp.ravel(x) for x in x_list])
        theta_vec = jnp.concatenate([jnp.ravel(x) for x in theta_list])
        theta0_vec = jnp.concatenate([jnp.ravel(x) for x in theta0])
        theta_final_vec = jnp.concatenate([jnp.ravel(x) for x in theta_final])

        step_size = (theta_final_vec - theta0_vec) / n_steps

        A = jax.jacobian(f_sys, 0)(x_vec, theta_vec)
        _, Bv = jax.jvp(lambda theta: f_sys(x_vec, theta), (theta_vec,), (step_size,))
        step = -jax.scipy.linalg.solve(A, Bv)

        x_next_vec = x_vec + step
        theta_next_vec = theta_vec + step_size

        x_next = flat_tensor_to_ragged_list(x_next_vec, x_shapes)
        theta_next = flat_tensor_to_ragged_list(theta_next_vec, theta_shapes)

        return [*x_next, *theta_next]

    f_step = jax.jit(step)
    return f_step


def compile_euler_approximation_function(A, B, variables, parameters, n_steps=100, mode=None):
    theta_final, result = euler_approximation(A, B, variables, parameters, n_steps=n_steps)
    theta_final.name = "theta_final"
    inputs = variables + parameters + [theta_final]

    f_euler = pytensor.function(inputs, result, mode=mode)
    f_euler.trust_inputs = True

    if mode in ["NUMBA", "JAX"]:
        _f_euler = f_euler.copy()

        def f_euler(*args, **kwargs):
            x = _f_euler.vm.jit_fn(*args, **kwargs)
            return list(map(np.asarray, x))

    return f_euler



import re

from collections.abc import Callable

import numba as nb
import numpy as np
import sympy as sp

from sympy.printing.numpy import NumPyPrinter, _known_functions_numpy

_known_functions_numpy.update({"DiracDelta": lambda x: 0.0, "log": "log"})


class NumbaFriendlyNumPyPrinter(NumPyPrinter):
    _kf = _known_functions_numpy

    def _print_Max(self, expr):
        # Use maximum instead of amax, because 1) we only expect scalars, and 2) numba doesn't accept amax
        return "{}({})".format(
            self._module_format(self._module + ".maximum"),
            ",".join(self._print(i) for i in expr.args),
        )

    def _print_Piecewise(self, expr):
        # Use the default python Piecewise instead of the numpy one -- looping with if conditions is faster in numba
        # anyway.
        result = []
        i = 0
        for arg in expr.args:
            e = arg.expr
            c = arg.cond
            if i == 0:
                result.append("(")
            result.append("(")
            result.append(self._print(sp.Float(e)))
            result.append(")")
            result.append(" if ")
            result.append(self._print(c))
            result.append(" else ")
            i += 1
        result = result[:-1]
        if result[-1] == "True":
            result = result[:-2]
            result.append(")")
        else:
            result.append(" else None)")
        return "".join(result)

    def _print_DiracDelta(self, expr):
        # The proper function should return infinity at one point, but the measure of that point is zero so this should
        # be fine. Pytensor defines grad(grad(max(0, x), x), x) to be zero everywhere.
        return "0.0"

    def _print_log(self, expr):
        return "{}({})".format(
            self._module_format(self._module + ".log"),
            ",".join(self._print(i) for i in expr.args),
        )


def numba_lambdify(
    exog_vars: list[sp.Symbol],
    expr: list[sp.Expr] | sp.Matrix | list[sp.Matrix],
    endog_vars: list[sp.Symbol] | None = None,
    func_signature: str | None = None,
    ravel_outputs=False,
) -> Callable:
    """
    Convert a sympy expression into a Numba-compiled function.  Unlike sp.lambdify, the resulting function can be
    pickled. In addition, common sub-expressions are gathered using sp.cse and assigned to local variables,
    giving a (very) modest performance boost. A signature can optionally be provided for numba.njit.

    Finally, the resulting function always returns a numpy array, rather than a list.

    Parameters
    ----------
    exog_vars: list of sympy.Symbol
        A list of "exogenous" variables. The distinction between "exogenous" and "enodgenous" is
        useful when passing the resulting function to a scipy.optimize optimizer. In this context, exogenous
        variables should be the choice varibles used to minimize the function.
    expr : list of sympy.Expr or sp.Matrix
        The sympy expression(s) to be converted. Expects a list of expressions (in the case that we're compiling a
        system to be stacked into a single output vector), a single matrix (which is returned as a single nd-array)
        or a list of matrices (which are returned as a list of nd-arrays)
    endog_vars : Optional, list of sympy.Symbol
        A list of "exogenous" variables, passed as a second argument to the function.
    func_signature: str
        A numba function signature, passed to the numba.njit decorator on the generated function.
    ravel_outputs: bool, default False
        If true, all outputs of the jitted function will be raveled before they are returned. This is useful for
        removing size-1 dimensions from sympy vectors.

    Returns
    -------
    numba.types.function
        A Numba-compiled function equivalent to the input expression.

    Notes
    -----
    The function returned by this function is pickleable.
    """
    ZERO_PATTERN = re.compile(r"(?<![\.\w])0([ ,\]])")
    FLOAT_SUBS = {
        sp.core.numbers.One(): sp.Float(1),
        sp.core.numbers.NegativeOne(): sp.Float(-1),
    }
    printer = NumbaFriendlyNumPyPrinter()

    if func_signature is None:
        decorator = "@nb.njit"
    else:
        decorator = f"@nb.njit({func_signature})"

    len_checks = []
    len_checks.append(
        f'    assert len(exog_inputs) == {len(exog_vars)}, "Expected {len(exog_vars)} exog_inputs"'
    )
    if endog_vars is not None:
        len_checks.append(
            f'    assert len(endog_inputs) == {len(endog_vars)}, "Expected {len(endog_vars)} exog_inputs"'
        )
    len_checks = "\n".join(len_checks)

    # Special case: expr is [[]]. This can occur if no user-defined steady-state values were provided.
    # It shouldn't happen otherwise.
    if expr == [[]]:
        sub_dict = ()
        code = ""
        retvals = ["[None]"]

    else:
        # Need to make the float substitutions so that numba can correctly interpret everything, but we have to handle
        # several cases:
        # Case 1: expr is just a single Sympy thing
        if isinstance(expr, sp.Matrix | sp.Expr):
            expr = expr.subs(FLOAT_SUBS)

        # Case 2: expr is a list. Items in the list are either lists of expressions (systems of equations),
        # single equations, or matrices.
        elif isinstance(expr, list):
            new_expr = []
            for item in expr:
                # Case 2a: It's a simple list of sympy things
                if isinstance(item, sp.Matrix | sp.Expr):
                    new_expr.append(item.subs(FLOAT_SUBS))
                # Case 2b: It's a system of equations, List[List[sp.Expr]]
                elif isinstance(item, list):
                    if all([isinstance(x, sp.Matrix | sp.Expr) for x in item]):
                        new_expr.append([x.subs(FLOAT_SUBS) for x in item])
                    else:
                        raise ValueError("Unexpected input type for expr")
            expr = new_expr
        else:
            raise ValueError("Unexpected input type for expr")
        sub_dict, expr = sp.cse(expr)

        # Converting matrices to a list of lists is convenient because NumPyPrinter() won't wrap them in np.array
        exprs = []
        for ex in expr:
            if hasattr(ex, "tolist"):
                exprs.append(ex.tolist())
            else:
                exprs.append(ex)

        codes = []
        retvals = []
        for i, expr in enumerate(exprs):
            code = printer.doprint(expr)

            delimiter = "]," if "]," in code else ","
            delimiter = ","
            code = code.split(delimiter)
            code = [" " * 8 + eq.strip() for eq in code]
            code = f"{delimiter}\n".join(code)
            code = code.replace("numpy.", "np.")

            # Handle conversion of 0 to 0.0
            code = re.sub(ZERO_PATTERN, r"0.0\g<1>", code)
            code_name = f"retval_{i}"
            retvals.append(code_name)
            code = f"    {code_name} = np.array(\n{code}\n    )"
            if ravel_outputs:
                code += ".ravel()"

            codes.append(code)
        code = "\n".join(codes)

    input_signature = "exog_inputs"
    unpacked_inputs = "\n".join(
        [
            f"    {getattr(x, 'safe_name', x.name)} = exog_inputs[{i}]"
            for i, x in enumerate(exog_vars)
        ]
    )
    if endog_vars is not None:
        input_signature += ", endog_inputs"
        exog_unpacked = "\n".join(
            [
                f"    {getattr(x, 'safe_name', x.name)} = endog_inputs[{i}]"
                for i, x in enumerate(endog_vars)
            ]
        )
        unpacked_inputs += "\n" + exog_unpacked

    assignments = "\n".join(
        [f"    {x} = {printer.doprint(y).replace('numpy.', 'np.')}" for x, y in sub_dict]
    )
    returns = f'[{",".join(retvals)}]' if len(retvals) > 1 else retvals[0]
    full_code = f"{decorator}\ndef f({input_signature}):\n{unpacked_inputs}\n\n{assignments}\n\n{code}\n\n    return {returns}"

    docstring = f"'''Automatically generated code:\n{full_code}'''"
    code = f"{decorator}\ndef f({input_signature}):\n    {docstring}\n{len_checks}\n{unpacked_inputs}\n\n{assignments}\n\n{code}\n\n    return {returns}"

    exec(code)
    return locals()["f"]


@nb.njit(cache=True)
def float_to_array(arr):
    return np.asarray(arr, np.float64)


@nb.njit(cache=True, nogil=True)
def euler_approx(f, x0, theta0, theta, n_steps, progress_bar):
    """
    Compute the solution to a non-linear function g(x, theta + dtheta) by iteratively computing a linear approximation
    f(x[t], theta + epsilon[t]) at the point (f(x[t-1], theta + epsilon[t-1]), theta + epsilon[t-1]), where epsilon[-1] = dtheta

    Parameters
    ----------
    f: njit function
        Linearized function to be approximated. Must have signature f(endog, exog) -> array[:]

    x0: np.ndarray
        Array of values of model variables representing the point at which g is linearized.

    theta0: np.ndarray
        Array of model parameter values representing the point at which g is linearized.

    theta: np.ndarray
        Values at which g is to be solved. These should correspond to something like "shocks" from the initial parameter
        values theta0.

    n_steps: int
        Number of gradient updates to perform; this is the length of the epsilon vector in the notation above. More steps
        leads to a more precise approximation.

    Returns
    -------
    x: np.ndarray
        Approximate solution to g(x + dx)

    Notes
    -----
    A non-linear function g(x, theta) = 0, can be linearized around a point (x0, theta0) as:

        A(x0, theta0) @ dx + B(x0, theta0) @ dtheta = 0

    Where A is the jacobian of dg/dx, and B is the jacobian dg/dtheta. This system can be solved for x:
        f(x0, theta0, dtheta) := dx = -inv(A(x0, theta0)) @ B(x0, theta0) @ dtheta

    It is well-known that this linear approximation is poor when dtheta is large relative to theta0. A
    solution to this problem is to decompse dtheta into a sequence of smaller -- presumably more accurate -- steps,
    and iteratively update [x0, theta0] in the following fashion:
        1. Initialize x_t = x0, theta_t = theta0
        2. Compute step_size = (theta - theta0) / n_steps
        3. For n_steps:
            1. Compute dx = f(x=x_t, theta=theta_t, dtheta=step_size)
            2. Update x_t = x_t + dx, theta_t = theta_t + step_size

    Using this algorithm, and given an infinite compute budget, g(x0, theta) can be computed to arbitrary precision.
    """
    x0 = np.atleast_1d(float_to_array(x0))
    theta0 = np.atleast_1d(float_to_array(theta0))
    theta = np.atleast_1d(float_to_array(theta))

    output = np.zeros((n_steps + 1, x0.size + theta0.size))
    output[0, : x0.size] = x0
    output[0, x0.size :] = theta0

    dtheta = theta - theta0
    x = np.concatenate((x0, theta0))
    step_size = dtheta / n_steps

    for t in range(1, n_steps + 1):
        dx = f(step_size, x).ravel()
        x = x + np.concatenate((dx, step_size))
        output[t, :] = x
        progress_bar.update(1)

    return output



from string import Template

import numpy as np
import pytest
import sympy as sp

from sympy.abc import a, b, x, y

from cge_modeling.base.primitives import Parameter, Variable
from cge_modeling.base.utilities import (
    _expand_var_by_index,
    flat_array_to_variable_dict,
    variable_dict_to_flat_array,
    wrap_fixed_values,
    wrap_pytensor_func_for_scipy,
)


def test_expand_variable():
    coords = {"i": ["A", "B", "C"]}
    x = Variable(name="x", dims="i", description="Sector <dim:i> demand for good <dim:j>")
    vars = _expand_var_by_index(x, coords)
    assert len(vars) == len(coords["i"])
    for variable, coord in zip(vars, coords["i"]):
        assert variable._full_latex_name == "x_{i=\\text{" + coord + "}}"
        assert variable.description == f"Sector {coord} demand for good <dim:j>"


def test_expand_variable_two_index():
    coords = {"i": ["A", "B", "C"], "j": ["A", "B", "C"]}
    x = Variable(name="x", dims="i, j", description="Sector i demand for good j")
    vars = _expand_var_by_index(x, coords)

    assert len(vars) == len(coords["i"]) * len(coords["j"])
    all_latex = [x._full_latex_name for x in vars]
    s = Template("x_{i=\\text{$i}, j=\\text{$j}}")
    for i in coords["i"]:
        for j in coords["j"]:
            assert s.substitute(i=i, j=j) in all_latex


def test_pack_and_unpack_is_bijective():
    variables = [
        Variable("Y", dims="i", description="<dim:i> output"),
    ]
    parameters = [
        Parameter(
            "phi",
            dims=("j", "i"),
            description="Input-output coefficient between <dim:i> and <dim:j>",
        ),
        Parameter("L_s", description="Total labor supply"),
    ]

    data_dict = {
        "Y": np.array([1, 2, 3]),
        "phi": np.array([[1, 2, 3], [4, 5, 6]]),
        "L_s": np.array(3.0),
    }

    coords = {"i": ["A", "B", "C"], "j": ["Q", "R"]}

    variable_array, param_array = variable_dict_to_flat_array(data_dict, variables, parameters)

    assert variable_array.shape[0] == sum(np.prod(data_dict[x.name].shape) for x in variables)
    assert param_array.shape[0] == sum(np.prod(data_dict[x.name].shape) for x in parameters)

    data_dict_2 = flat_array_to_variable_dict(
        np.concatenate([variable_array, param_array]), variables + parameters, coords
    )

    assert all(
        [data_dict[x.name].shape == data_dict_2[x.name].shape for x in variables + parameters]
    )
    assert all(
        [np.allclose(data_dict[x.name], data_dict_2[x.name]) for x in variables + parameters]
    )


def test_wrap_pytensor_func_for_scipy():
    import pytensor
    import pytensor.tensor as pt

    variables = [Variable(name, dims=None) for name in ["x", "y", "z"]]
    parameters = [Parameter(name, dims=None) for name in ["a", "b", "c"]]
    coords = {}

    x, y, z = (pt.dscalar(var.name) for var in variables)
    a, b, c = (pt.dscalar(var.name) for var in parameters)

    f = a * x + b * y + c * z
    mse = (f**2).sum()

    f_mse = pytensor.function([x, y, z, a, b, c], mse, mode="FAST_COMPILE")
    test_value_dict = {"a": 1.0, "b": 2.0, "c": 3.0, "x": 1.0, "y": 2.0, "z": 3.0}
    assert f_mse(**test_value_dict) == 196.0

    f_mse_wrapped = wrap_pytensor_func_for_scipy(f_mse, variables, parameters, coords)
    assert f_mse_wrapped(np.array([1.0, 2.0, 3.0]), np.array([1.0, 2.0, 3.0])) == 196.0


def test_fixed_values_wrapper():
    variables = [
        Variable("x", dims=None, description="x"),
        Variable("y", dims=None, description="y"),
    ]

    eq = a * x + b * y
    f = sp.lambdify([x, y, a, b], eq)
    x_val = np.random.normal(scale=10)
    f = wrap_fixed_values(f, {"x": x_val}, variables, coords={})

    assert f(y=1, a=1, b=1) == (x_val + 1)


def test_fixed_values_wrapper_raises_on_unknown_variable():
    variables = [
        Variable("x", dims=None, description="x"),
        Variable("y", dims=None, description="y"),
    ]

    eq = a * x + b * y
    f = sp.lambdify([x, y, a, b], eq)

    with pytest.raises(ValueError):
        f = wrap_fixed_values(f, {"z": 2}, variables, coords={})



import pytest

from cge_modeling.base.primitives import Parameter, Variable


@pytest.mark.parametrize("cls", [Variable, Parameter], ids=["variable", "parameter"])
@pytest.mark.parametrize(
    "name, latex_name, dims, expected",
    [
        ("x", "x", None, ("x", "x, Positive = True, Real = True", ())),
        ("x", "x", ("i",), ("x_{i}", "x_{i}, Positive = True, Real = True", ("i",))),
        (
            "x",
            r"\Omega",
            "i",
            (r"\Omega_{i}", r"\Omega_{i}, Positive = True, Real = True", ("i",)),
        ),
        (
            "x",
            "x",
            "i,j",
            ("x_{i, j}", "x_{i, j}, Positive = True, Real = True", ("i", "j")),
        ),
        (
            "x_d",
            "x_d",
            "i,j",
            ("x_{d, i, j}", "x_{d, i, j}, Positive = True, Real = True", ("i", "j")),
        ),
    ],
    ids=[
        "No_dims",
        "tuple_dims",
        "single_string",
        "double_string",
        "underscore_name",
    ],
)
def test_create_variable_defaults(cls, name, latex_name, dims, expected):
    # noinspection PyArgumentList
    x = cls(name=name, latex_name=latex_name, dims=dims, extend_subscript=True)
    expected_latex_name, description, dims = expected

    assert x._full_latex_name == expected_latex_name
    assert x.description == description
    assert x.dims == dims


@pytest.mark.parametrize("cls", [Variable, Parameter], ids=["variable", "parameter"])
def test_create_variable(cls):
    # noinspection PyArgumentList
    x = cls(name="x", dims="i", description="A lovely variable")

    # noinspection PyArgumentList
    y = cls(
        name="y",
        dims="i",
        description="Another lovely variable",
    )

    assert x._full_latex_name == r"x_{i}"
    assert x.description == "A lovely variable"
    assert x.dims == ("i",)
    assert y > x


@pytest.mark.parametrize("cls", [Variable, Parameter], ids=["variable", "parameter"])
def test_to_dict(cls):
    # noinspection PyArgumentList
    x = cls(name="x", dims="i", description="A lovely thing")
    d = x.to_dict()

    assert isinstance(d, dict)
    assert len(d) == 6
    assert all([key in d.keys() for key in ["name", "dims", "dim_vals", "description"]])


@pytest.mark.parametrize(
    "base_name, dims, extend, expected",
    [
        ("x", ["i"], False, "x_{i=\\text{A}}"),
        ("x^j", ["i"], False, "x^j_{i=\\text{A}}"),
        ("x^{j}", ["i"], False, "x^{j}_{i=\\text{A}}"),
        ("x_F", ["i", "j"], True, "x_{F, i=\\text{A}, j}"),
        ("x_{Fish}", ["i", "i"], True, "x_{Fish, i=\\text{A}, i=\\text{A}}"),
        ("x_K_d", ["i", "j"], 2, "x_{K, d, i=\\text{A}, j}"),
        ("var_with_underscore", ["i"], False, "var_with_underscore_{i=\\text{A}}"),
    ],
)
def test_sub_label(base_name, dims, extend, expected):
    x = Variable(
        name=base_name,
        dims=dims,
        description="A lovely item from group <dim:i>",
        extend_subscript=extend,
    )
    x.update_dim_value("i", "A")
    assert x._full_latex_name == expected


@pytest.mark.parametrize(
    "description, expected",
    [
        ("A lovely item from group <dim:i>", "A lovely item from group A"),
        ("Sector <dim:i> demand for good <dim:j>", "Sector A demand for good <dim:j>"),
        ("<dim:i> demand for <dim:j>", "A demand for <dim:j>"),
    ],
)
def test_sub_description(description, expected):
    x = Variable(name="x", dims=["i", "j"], description=description)
    x.update_dim_value("i", "A")
    assert x.description == expected
